# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Test the `basic` module."""

import numpy as np

from pylook.calc import (elastic_correction, friction, remove_offset, zero)
from pylook.testing import assert_array_almost_equal
from pylook.units import units


def test_zero_defaults():
    """Test zero with all of the default args."""
    data = np.arange(10) * units('mm')

    result = zero(data, 5)

    truth = np.array([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_zero_mode_before():
    """Test zero with the mode of before."""
    data = np.arange(10) * units('mm')

    result = zero(data, 5, mode='before')

    truth = np.array([0, 0, 0, 0, 0, 0, 1, 2, 3, 4]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_zero_mode_after():
    """Test zero with the mode of after."""
    data = np.arange(10) * units('mm')

    result = zero(data, 5, mode='after')

    truth = np.array([-5, -4, -3, -2, -1, 0, 0, 0, 0, 0]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_zero_window():
    """Test zero with a window to get an average zero value."""
    data = np.array([0, 1, 2, 2.2, 2.5, 2.3, 2.2, 2.6, 2.7, 2.9, 3]) * units('mm')

    result = zero(data, 5, window=2)

    truth = data - 2.36 * units('mm')

    assert_array_almost_equal(result, truth)


def test_zero_value_at_mode():
    """Test zeroing with an offset value in the default at mode."""
    data = np.arange(10) * units('mm')

    result = zero(data, 5, value=1.5 * units('mm'))

    truth = np.array([-3.5, -2.5, -1.5, -0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_zero_value_before_mode():
    """Test zeroing with an offset value in the before mode."""
    data = np.arange(10) * units('mm')

    result = zero(data, 5, value=1.5 * units('mm'), mode='before')

    truth = np.array([1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 2.5, 3.5, 4.5, 5.5]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_zero_value_after_mode():
    """Test zeroing with an offset value in the after mode."""
    data = np.arange(10) * units('mm')

    result = zero(data, 5, value=1.5 * units('mm'), mode='after')

    truth = np.array([-3.5, -2.5, -1.5, -0.5, 0.5, 1.5, 1.5, 1.5, 1.5, 1.5]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_remove_offset():
    """Test the remove offset function."""
    data = np.array([0, 1, 2, 4, 4, 10, 10, 11, 12, 13, 14]) * units('mm')

    result = remove_offset(data, 4, 6)

    truth = np.array([0, 1, 2, 4, 4, 10, 4, 5, 6, 7, 8]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_remove_offset_set_between():
    """Test the remove offset function."""
    data = np.array([0, 1, 2, 4, 4, 10, 10, 11, 12, 13, 14]) * units('mm')

    result = remove_offset(data, 4, 6, set_between=True)

    truth = np.array([0, 1, 2, 4, 4, 4, 4, 5, 6, 7, 8]) * units('mm')

    assert_array_almost_equal(result, truth)


def test_elastic_correction_linear_same_units():
    """Test the elastic correction with all consistent units given."""
    coeffs = [5 * units('mm/kN'), 10 * units('mm')]
    loads = np.arange(10, 101, 10) * units('kN')
    displacements = (np.arange(1, 11) * 1000) * units('mm')

    result = elastic_correction(loads, displacements, coeffs)

    truth = np.array([940, 1890, 2840, 3790, 4740, 5690, 6640, 7590, 8540, 9490]) * units('mm')
    assert_array_almost_equal(result, truth)


def test_elastic_correction_linear_different_units():
    """Test the elastic correction with inconsistent units given."""
    coeffs = [5 * units('mm/kN'), 10 * units('mm')]
    loads = (np.arange(10, 101, 10) * 1000) * units('N')
    displacements = (np.arange(1, 11) * 1000000) * units('micron')

    result = elastic_correction(loads, displacements, coeffs)

    truth = np.array([940000, 1890000, 2840000, 3790000, 4740000,
                      5690000, 6640000, 7590000, 8540000, 9490000]) * units('micron')
    assert_array_almost_equal(result, truth)


def test_elastic_correction_quadratic_same_units():
    """Test the elastic correction with all consistent units given."""
    coeffs = [2 * units('mm/kN**2'), 5 * units('mm/kN'), 10 * units('mm')]
    loads = np.arange(10, 101, 10) * units('kN')
    displacements = (np.arange(1, 11) * 1000) * units('mm')

    result = elastic_correction(loads, displacements, coeffs)

    truth = np.array([740, 1090, 1040, 590, -260, -1510, -3160, -5210,
                      -7660, -10510]) * units('mm')
    assert_array_almost_equal(result, truth)


def test_friction_different_units():
    """Test that the friction calculation produces expected values with different units."""
    sigma_n = np.array([-100, 0, 1000, 2000, 3000, 4000, 5000]) * units('N')
    tau = np.array([0, 0, 2.2, 2.2, 2.2, 2.2, 2.2]) * units('kN')

    result = friction(tau, sigma_n)

    truth = np.array([0, 0, 2.2, 1.1, 0.7333333, 0.55, 0.44]) * units('dimensionless')

    assert_array_almost_equal(result, truth)


def test_friction():
    """Test that the friction calculation produces expected values with same units."""
    sigma_n = np.array([-.1, 0, 1, 2, 3, 4, 5]) * units('kN')
    tau = np.array([0, 0, 2.2, 2.2, 2.2, 2.2, 2.2]) * units('kN')

    result = friction(tau, sigma_n)

    truth = np.array([0, 0, 2.2, 1.1, 0.7333333, 0.55, 0.44]) * units('dimensionless')

    assert_array_almost_equal(result, truth)



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Test the `lookfiles` module."""

from pylook.io import (XlookParser)


def test_look_parser_creation():
    """Make sure we can create an empty instance of the look r file parser."""
    XlookParser()



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Contains basic calculations needed when processing data."""

import numpy as np

from ..package_tools import Exporter

exporter = Exporter(globals())


@exporter.export
def zero(data, zero_idx, window=0, value=0, mode='at'):
    """
    Zero an array at a given row to zero or another value.

    Parameters
    ----------
    data : `pint.Quantity`
        Data to be operated upon.
    zero_idx : int
        Index of value at which we wish to zero the array
    window : int
        Number of data points either side of the zero index to be averaged to get the
        zero value.
    value : `pint.Quantity`
        Numeric value to which we set the "zero" point.
    mode : string
        How the zero operation should be performed. Valid modes at "at" in which only the
        zero value is subtracted from all data, "before" in which all data before that index
        are also set to the zero value, or "after" in which all data after the given index
        are also set to the zero value.

    Returns
    -------
    data : `pint.Quantity`
        Data with zero applied.
    """
    # First we get the value we are going to use as zero - a single value or a mean
    if window:
        zero_value = np.mean(data[zero_idx - window: zero_idx + window + 1])
    else:
        zero_value = data[zero_idx]

    # Zero the data to that value
    data = data - zero_value

    # If there is a value we want to set the data to, add it
    if value:
        data = data + value

    # If the mode is before/after we need to zero out those values
    if mode == 'before':
        data[0:zero_idx] = data[zero_idx]

    if mode == 'after':
        data[zero_idx:] = data[zero_idx]

    return data


@exporter.export
def remove_offset(data, start_idx, end_idx, set_between=False):
    """
    Remove offsets in the data.

    Parameters
    ----------
    data : `pint.Quantity`
        Date to be operated upon.
    start_idx : int
        Index that marks the start of the offset.
    end_idx : int
        Index that marks the end of the offset.
    set_between : int
        Set the data after the start point up to the end point to have the
        value of the start point. Default is `False`.

    Returns
    -------
    data : `pint.Quantity`
        Data with offset applied.
    """
    # Set the data after the offset to the data minus the offset
    offset = data[end_idx] - data[start_idx]

    # Set the intemediate data (during the offset) to be the first
    # value if so desired.
    if set_between:
        data[start_idx: end_idx] = data[start_idx]

    data[end_idx:] = data[end_idx:] - offset
    return data


@exporter.export
def elastic_correction(load, displacement, coeffs):
    """
    Perform an elastic correction on a single axis of data.

    Parameters
    ----------
    load : `pint.Quantity`
        Load/Force data
    displacement : `pint.Quantity`
        Displacement data
    coeffs : list
        list of coefficients from highest power to lowest.

    Returns
    -------
    displacement : `pint.Quantity`
        Displacement with the elastic correction applied
    """
    # store in incoming displacement unit
    displacement_units_incoming = displacement.units

    # convert everythign to base units, then drop them
    load = load.to_base_units().m
    displacement = displacement.to_base_units()
    displacement_base_units = displacement.units
    displacement = displacement.m
    coeffs = [c.to_base_units().m for c in coeffs]

    # Find the elastic correction
    elastic_correction = np.polyval(coeffs, load)

    # Determine the final elastic corrected displacement and attach the base units
    elastic_corrected_displacement = ((displacement - elastic_correction)
                                      * displacement_base_units)

    return elastic_corrected_displacement.to(displacement_units_incoming)


@exporter.export
def friction(shear_component, normal_component):
    """
    Calculate the simple friction.

    Parameters
    ----------
    shear_component : `pint.Quantity`
        Shear component of load/force/stress
    normal_component : `pint.Quantity`
        Normal component of load/force/stress

    Returns
    -------
    friction : `pint.Quantity`
        Simple friction value

    Notes
    -----
    Modifies the normal load/force/ stress to have a minimum value of 1e-16 to avoid any divide
    by zero warnings or negative friction values due to the normal component.
    """
    # Clip the normal component to always be slightly above zero
    normal_component = np.clip(normal_component, 1e-16 * normal_component.units, None)
    return shear_component / normal_component



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""
============================
Reducing p655 in Pure Python
============================

Experiment p655 was run in 2005 by Mckiernan, Rathbun, and Rowe. It was reduced by
Chris Marone. Here we use the xlook r file parser to run that r file and then get it into the
dictionary of quantity arrays like we use everywhere else and do some simple plotting
of the experiment.

For the curious, this experiment is determining the frictional response of "Ghost Rocks"
from Kodiak Alaska.

While being able to run r files is great, we'd like to move to reducing the data in Python
to really get at the full power of pylook. There are lots of ways that pylook can simplify
reduction and it is easier to follow exactly what is happening for those unfamiliar with the
r file format.

Let's start out with some imports to get rolling.
"""

import numpy as np

import pylook.calc as lc
from pylook.cbook import get_test_data
from pylook.io import read_binary
from pylook.units import units

##############################

data_path = get_test_data('p655intact100l')

##############################

data, _ = read_binary(data_path)

##############################
# The time column in these data files is really the sample rate in Hz. We want to turn that
# into a delta time and cumulatively sum to get experiment elapsed time. Notice that we are
# assigning units by multiplying them - easy! We have to take the magnitude of time for
# `cumsum` because numpy silently drops our units. We'll fix that with a `cumsum` wrapper
# in pylook soon.

data['Time'] = np.cumsum(1 / data['Time'].m) * units('s')

##############################
# Now we need to apply calibrations and units to the data - the calibrations are determined
# based on the sensor used by the experimentalist. We'll get a list of the current data column
# names so we know what they are called in the look file.

data.keys()

##############################

data['Vert_Disp'] = data['Vert_Disp'] * 0.076472 * units('micron / bit')
data['Vert_Load'] = data['Vert_Load'] * 1.597778959e-3 * units('MPa / bit')
data['Hor_Disp'] = data['Hor_Disp'] * 0.11017176 * units('micron / bit')
data['Hor_Load.'] = data['Hor_Load.'] * 3.31712805707e-3 * units('MPa / bit')

##############################

# Now that calibrations are applied, we really should rename those columns to something more
# useful. Look limited the length of column names, but pylook does not. We even allow spaces!
# The `pop` method removes that key/value pair from the dictionary and we reassign it to a
# new key.

data['Shear Displacement'] = data.pop('Vert_Disp')
data['Shear Stress'] = data.pop('Vert_Load')
data['Normal Displacement'] = data.pop('Hor_Disp')
data['Normal Stress'] = data.pop('Hor_Load.')

##############################
# There was a spike in the data that the experimentalist decided to remove. We can do that
# with an offset that sets everything in between the rows to the final value.

data['Shear Stress'] = lc.remove_offset(data['Shear Stress'], 4075, 4089, set_between=True)

##############################
# The testing machine expands when loading because it has a finite stiffness. That stiffness
# is known and can be removed from the data to create an elastically corrected displacement.
# Python is great with units here and can help you calculate the stiffness in the right units.
# We know the stiffness is 0.37 kN/micron and this experiment use samples that were 5x5cm.
# From that we let the units library do all of the hard work of calculating the stiffness in
# terms of stress.

normal_stiffness = 0.37 * units('kN/micron')
sample_normal_area = 5 * units('cm') * 5 * units('cm')
normal_stiffness = sample_normal_area / normal_stiffness
print(normal_stiffness.to('micron/MPa'))  # To just displays things in units that we expect

##############################
# pylook's elastic correction is a polynomial tool, so you can use as many coefficients as
# you'd like. Just like `np.polyval` we expect coefficients from the highest order to lowest.
# Units matter!

data['Normal Displacement'] = lc.elastic_correction(data['Normal Stress'],
                                                    data['Normal Displacement'],
                                                    [normal_stiffness, 0 * units('micron')])

##############################
# We start recording data before there is any load on the samples, so we need to find where
# the load is brought on. Easiest way to do this is to make a quick plot and find the row
# number at which we want to zero things. The `.m` after the data is a way to drop units
# (the magnitude) and is required as currently bokeh does not play well with units.

from bokeh.io import output_notebook
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show

output_notebook()

##############################

p = figure(title='Find the normal stress zero row', tools='box_zoom, reset, hover')
p.line(data['rec_num'].m, data['Normal Stress'].m)
show(p)

##############################
# Row 42 looks pretty good, so we zero the normal stress there. We'd also like to set
# everything before that row to zero since it's just noise. In r files that took some math
# commands, we zero has options in pylook! We also don't need to worry about adding small
# values to avoid divide by zero errors as the friction calculation handles that properly.

data['Normal Stress'] = lc.zero(data['Normal Stress'], 42, mode='before')

##############################
# While we are zeroing, it is a good time to deal with the normal displacement as well.

data['Normal Displacement'] = lc.remove_offset(data['Normal Displacement'], 0, 42,
                                               set_between=True)

##############################
# Now we need to find the zero point for the shear load and stress.

p = figure(title='Find the shear stress zero row', tools='box_zoom, reset, hover')
p.line(data['rec_num'].m, data['Shear Stress'].m)
show(p)

##############################
# Row 1518 looks like a reasonable choice - so lets' zero those two and zero everything
# before then.

data['Shear Displacement'] = lc.zero(data['Shear Displacement'], 1518, mode='before')
data['Shear Stress'] = lc.zero(data['Shear Stress'], 1518, mode='before')

##############################
# The displacement transducer on the shear axis often runs out of travel and has to be reset
# during the experiment. We need to find those spots and record the rows at which those
# offsets start and stop so we can remove them.

p = figure(title='Find the DCDT offsets', tools='box_zoom, reset, hover')
p.line(data['rec_num'].m, data['Shear Displacement'].m)
show(p)

##############################
# Looks like we had two offsets - rows 18593 to 19058 and rows 66262 to 67830.
# Let's remove those and set the values between to the final value to that data look nice.

data['Shear Displacement'] = lc.remove_offset(data['Shear Displacement'],
                                              18593, 19058, set_between=True)
data['Shear Displacement'] = lc.remove_offset(data['Shear Displacement'],
                                              66262, 67830, set_between=True)

##############################
# For the normal displacement we assume that half of it is in each of the two layers of a
# double direct setup. We are going to change the sign such that compaction shows a thinner
# layer and dialation shows a thicker layer. On the bench the blocks were 89.2 mm thick with
# a 4 mm gouge layer. Notice that because of the units awareness we can just add 4 mm and the
# math works!

data['Normal Displacement'] = data['Normal Displacement'] * (-0.5 * units('dimensionless'))
data['Normal Displacement'] = lc.zero(data['Normal Displacement'], 42, mode='before')
data['Normal Displacement'] = data['Normal Displacement'] + 4 * units('mm')

##############################
# Let's calculate simple friction (shear stress / normal stress). We have a function in pylook
# that does this intelligently handling divide by zeros when there is no shear stress on the
# sample

data['Friction'] = lc.friction(data['Shear Stress'], data['Normal Stress'])

##############################
# ## Quick Look
# We'll use Bokeh to take a quick look at the data. Matplotlib is the best choice for your
# publication plots, but the speed and interactivity of Bokeh in the notebook is hard to beat.
# We'll be adding helpers to pylook to make this process easier in the future as well.

# This is a handy function that will be integrated into pylook in a more advanced way soon,
# but demonstrates how to make a flexible plotting function instead of copying and pasting a
# bunch of code over and over again.


def make_runplot(data, x_var='Time', y_vars=None,
                 tools='pan,wheel_zoom,box_zoom,reset,save,box_select,hover'):
    plots = []
    for col_name in list(data):
        if col_name == x_var:
            continue
        if y_vars and (col_name not in y_vars):
            continue

        # First plot is simple, the rest we share the x range with the first
        if plots == []:
            p = figure(title=col_name, tools=tools)
        else:
            p = figure(title=col_name, tools=tools, x_range=plots[0].x_range)

        # Plot the data and set the labels
        p.xaxis.axis_label = str(data[x_var].units)
        p.yaxis.axis_label = str(data[col_name].units)
        p.line(data[x_var].m, data[col_name].m)

        plots.append(p)
    show(gridplot(plots, ncols=1, plot_width=600, plot_height=175))

##############################
# By default make_runplot would plot all of the variables, let's just plot a couple of basic
# ones. Hover over the graph to see the values! That can be turned off by clicking the message
# bubble icon in the plot toolbar. If we don't specify, data are plotted with respect to time.


make_runplot(data, y_vars=['Shear Stress', 'Normal Stress'])


##############################
# We can specify to plot relative to another x variable though - with load point displacement
# probably being the most common.

make_runplot(data, x_var='Shear Displacement', y_vars=['Shear Stress', 'Normal Stress'])

##############################
# Easy right? We'll be adding more features to reduce this code even further, but it moves us
# even further towards the goal of easy to understand and portable reductions that require a
# minimum of installation and learning pain. Being in Python also means we can seamlessly
# transfer these data into machine learning tools or just about any other analysis library
# you're interested in!



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Tools for versioning."""


def get_version():
    """Get the package version.

    Either get it from package metadata, or get it using version control information if
    a development install.
    """
    try:
        from setuptools_scm import get_version
        return get_version(root='..', relative_to=__file__,
                           version_scheme='post-release', local_scheme='dirty-tag')
    except (ImportError, LookupError):
        from pkg_resources import get_distribution
        return get_distribution(__package__).version



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Provide calculations and data processing aids."""

from .lookfiles import *  # noqa: F403

__all__ = lookfiles.__all__[:]  # noqa: F405



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Contains utilities to work with "look" style data files and associated "r" files."""

from pathlib import Path
import struct
import warnings

import numpy as np
from pint.errors import UndefinedUnitError

import pylook.calc as lc
from pylook.units import units
from ..package_tools import Exporter

exporter = Exporter(globals())


def _binary_tuple_to_string(binary_form):
    out = []
    for c in binary_form:
        # Stop at null termination
        if c == b'\x00':
            break
        out.append(c.decode())
    return ''.join(out)


@exporter.export
def read_binary(filename, data_endianness=None, unrecognized_units='ignore',
                clean_header=True):
    """
    Read a look binary formatted file into a dictionary of united arrays.

    Parameters
    ----------
    filename : string
        Filename or path to file to read
    data_endianness: string
        Endianness of the data section of the file. None, 'big', or 'little'.
        None interprets the file as it believes fit, big and little force the
        endianness.
    unrecogized_units : string
        'ignore' (defualt) assigns dimensionless to unrecognized units, 'error' will
        fail if unrecognized units are encountered.
    clean_header : boolean
        Remove extra whitespace in the header data column names and units. Default True.

    Returns
    -------
    data : dict
        Dictionary of `pint.Quantity` arrays for each column of data.
    metadata : dict
        Metadata from the header of the file

    Notes
    -----
    The data section of the file is written in the native format of the machine
    used to produce the file.  Endianness of data is little by default, but may
    be changed to 'big' to accomodate older files or files written on power pc
    chips.
    """
    if type(filename) == str:
        filename = Path(filename)

    if data_endianness is None:
        data_endianness = 'little'

    metadata = _read_binary_file_metadata(filename, clean_header=clean_header)

    if metadata['bytes per data point'] == 4:
        data_endianness = 'big'

    with open(filename, 'rb') as f:

        # Seek past the file metadata header that we have already processed
        f.seek(36)

        col_headings = []
        col_recs = []
        col_units = []

        # For each possible column (32 maximum columns) unpack its header
        # information and store it.  Only store column headers of columns
        # that contain data.  Use termination at first NULL.
        for i in range(metadata['header format']):
            # Channel name (13 characters)
            chname = struct.unpack('13c', f.read(13))
            chname = _binary_tuple_to_string(chname)
            chname = chname.split('\0')[0]

            # Channel units (13 characters)
            chunits = struct.unpack('13c', f.read(13))
            chunits = _binary_tuple_to_string(chunits)
            chunits = chunits.split('\0')[0]

            # This field is now unused, so we just read past it (int)
            _ = struct.unpack('>i', f.read(4))

            # This field is now unused, so we just read past it (50 characters)
            _ = struct.unpack('50c', f.read(50))

            # Number of elements (int)
            nelem = struct.unpack('>i', f.read(4))
            nelem = int(nelem[0])

            if clean_header:
                chname = chname.strip()
                chunits = chunits.strip()

            if chname[0:6] == 'no_val':
                continue  # Skip Blank Channels
            else:
                col_headings.append(chname)
                col_recs.append(nelem)
                col_units.append(chunits)

        # Read the data into a numpy recarray
        data = np.empty([metadata['number of records'], metadata['number of columns']])

        # Make the right data formatter for the file
        if metadata['bytes per data point'] == 8:
            data_point_format_little_endian = '<d'
            data_point_format_big_endian = '>d'
        elif metadata['bytes per data point'] == 4:
            data_point_format_little_endian = '<f'
            data_point_format_big_endian = '>f'
        else:
            ValueError('Bytes per data must be 4 or 8. Got'
                       f" {metadata['bytes per data point']}")

        for col in range(metadata['number of columns']):
            for row in range(col_recs[col]):
                if data_endianness == 'little':
                    data[row, col] = struct.unpack(data_point_format_little_endian,
                                                   f.read(metadata['bytes per data point']))[0]
                elif data_endianness == 'big':
                    data[row, col] = struct.unpack(data_point_format_big_endian,
                                                   f.read(metadata['bytes per data point']))[0]
                else:
                    ValueError('Data endian setting invalid - options are little and big')

    data_dict = {}
    data_dict['rec_num'] = np.arange(metadata['number of records']) * units('dimensionless')

    for i, (name, unit) in enumerate(zip(col_headings, col_units)):
        data_unit = units('dimensionless')
        try:
            data_unit = units(unit)

        except UndefinedUnitError:
            if unrecognized_units == 'ignore':
                warnings.warn(f'Unknown unit {unit} - assigning dimensionless units.')
            else:
                raise UndefinedUnitError(unit)

        data_dict[name] = data[:, i] * data_unit

    return data_dict, metadata


def _determine_header_and_data_format(file_size, num_channels, num_records):
    """
    Determine the column metadata header and data format of the file.

    Look files have been through several generations as data and systems have expanded.
    This helper determines if the file was written to hold up to 16 or 32 columns and
    if the data are stored as 4 byte floats or 8 byte doubles.

    Parameters
    ----------
    file_size : int
        Total file size in bytes
    num_channels : int
        Number of data channels written to the file
    num_records : int
        Number of records (rows) written to each channel

    Returns
    -------
    number_of_header_channels : int
        Number of channels in the header, 16 or 32.
    bytes_per_data : int
        Nubmer of bytes per data point, 4 or 8.
    """
    # Calculate the size if this were a 16 channel file - these files used all 4 byte floats
    sixteen_ch_float_file_size = 36 + 84 * 16 + 4 * num_records * num_channels

    # Calculate the size if this were a 32 channel file of 4 byte floats
    thirty_two_ch_float_file_size = 36 + 84 * 32 + 4 * num_records * num_channels

    # Calculate the size if this were a 32 channel file of 8 byte doubles
    thirty_two_ch_double_file_size = 36 + 84 * 32 + 8 * num_records * num_channels

    if file_size == sixteen_ch_float_file_size:
        return 16, 4
    elif file_size == thirty_two_ch_float_file_size:
        return 32, 4
    elif file_size == thirty_two_ch_double_file_size:
        return 32, 8
    else:
        IOError(f'Cannot determine format of look file with size {file_size}')


def _read_binary_file_metadata(filename, clean_header=True):
    """
    Read the file metadata and detemine the file format of a look file.

    Parameters
    ----------
    filename : string
        Filename or path to file to read
    clean_header : boolean
        Remove extra whitespace in the header data column names and units. Default True.

    Returns
    -------
    metadata : dict
        Dictionary of file metadata
    """
    with open(filename, 'rb') as f:
        metadata = {}

        # Unpack information at the top of the file about the experiment
        name = struct.unpack('20c', f.read(20))
        name = _binary_tuple_to_string(name)
        name = name.split('\0')[0]

        if clean_header:
            name = name.strip()
            metadata['name'] = name
        # The rest of the header information is written in big endian format

        # Number of records (int)
        num_recs = struct.unpack('>i', f.read(4))
        num_recs = int(num_recs[0])
        metadata['number of records'] = num_recs

        # Number of columns (int)
        num_cols = struct.unpack('>i', f.read(4))
        num_cols = int(num_cols[0])
        metadata['number of columns'] = num_cols

        # Sweep (int) - No longer used
        swp = struct.unpack('>i', f.read(4))[0]
        metadata['swp'] = swp

        # Date/time(int) - No longer used
        dtime = struct.unpack('>i', f.read(4))[0]
        metadata['dtime'] = dtime

        # Get the total size of the file
        metadata['file size'] = filename.stat().st_size

        # Determine the type of column header and data we're going to
        # encounter and add that to metadata
        res = _determine_header_and_data_format(metadata['file size'],
                                                metadata['number of columns'],
                                                metadata['number of records'])
        metadata['header format'], metadata['bytes per data point'] = res

        return metadata


@exporter.export
class XlookParser:
    """Xlook R File Parser."""

    def __init__(self):
        """Initialize the parser with 32 empty columns to match the Look data model."""
        self.max_cols = 32
        self.data = [np.array([]) for i in range(self.max_cols)]
        self.data_units = [None for i in range(self.max_cols)]
        self.data_names = [None for i in range(self.max_cols)]

    def _get_data_by_index(self, index):
        """
        Get a data array by its zero based index column number.

        Parameters
        ----------
        index : int
            Index (zero based) of the data column to get.

        Returns
        -------
        data : `pint.Quantity`
            Data array.
        """
        return self.data[index]

    def _get_units_by_index(self, index):
        """
        Get the units associated with the zero based column number.

        Parameters
        ----------
        index : int
            Index (zero based) of the data column unit to get

        Returns
        -------
        unit : str
            String representation of the column units
        """
        return self.data_units[index]

    def _get_name_by_index(self, index):
        """
        Get the name for a column associated with the zero based column number.

        Parameters
        ----------
        index : int
            Index (zero based) of the data column name to get

        Returns
        -------
        name : str
            Name of the data column
        """
        return self.data_names[index]

    def _set_units_by_index(self, index, units):
        """
        Store the units associated with a column based on its zero based column number.

        Parameters
        ----------
        index : int
            Index (zero based) of the data column unit to set
        units : str
            Unit name to associate with the data column
        """
        self.data_units[index] = units

    def _set_name_by_index(self, index, name):
        """
        Set the name associated with a column based on its zero based column number.

        Parameters
        ----------
        index : int
            Index (zero based) of the data column name to set
        name : str
            Name to associate with the data column
        """
        self.data_names[index] = name

    def _set_data_by_index(self, index, data):
        """
        Store data into a given zero based index position.

        Parameters
        ----------
        index : int
            Index (zero based) of the data column to set
        data : `pint.Quantity`
            Data to associate with the data column
        """
        self.data[index] = data

    def doit(self, rfile, endianness=None):
        """
        Run an r-file - naming directly from XLook itself for ease of learning for new users.

        Parameters
        ----------
        rfile : str
            Path to r file to run
        endianness: str
            None, little, or big. Defaults to None which lets the reader try to determine this,
            but can be forced if needed.
        """
        with open(rfile, 'r') as f:
            self._r_file_path = Path(rfile)
            for line in f.readlines():
                # If there is an in-line comment, we split and just keep the first part
                if '#' in line:
                    line = line.split('#')[0].strip()

                # If it is an end command - bounce out of doit
                if line.strip() == 'end':
                    return
                self.parse_line(line)

    def parse_line(self, line):
        """
        Parse the text in an xlook command and execute the appropriate function.

        Parameters
        ----------
        line : str
            Xlook command line to process
        """
        # Kill any trailing whitespaces
        line = line.strip()

        # There doesn't have to be a space after the #, so let's just short circuit
        # here for any line starting with a #.
        if (line.startswith('#')) or (line.strip() == ''):
            self.command_comment(line)
            return

        # Split up the command into the root and the rest
        command_root = line.strip().split(' ')[0]

        # There were never supposed to be commas in the lines, but some files
        # had them we replace them with a space for arg seperation.
        line = line.replace(',', ' ')

        # Dictionary mapping commands to their functions
        command_functions = {'math': self.command_math,
                             'begin': self.command_begin,
                             'com_file': self.command_com_file,
                             'summation': self.command_summation,
                             'power': self.command_power,
                             'zero': self.command_zero,
                             'ec': self.command_ec,
                             'r_col': self.command_r_col,
                             'math_int': self.command_math_int,
                             'offset_int': self.command_offset_int,
                             'r_row': self.command_r_row,
                             #  'type': self.command_type,
                             'read': self.command_read}

        # Determine the function that we should run - if there isn't a matching function
        # we issue a warning and keep running. That's what xLook did.
        action_function = command_functions.get(command_root)
        if action_function is None:
            action_function = self.command_invalid

        action_function(line.strip())

    def command_math(self, command):
        r"""
        Perform math operations with columns of data.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The XLook command is `math x_col_number operator y_col_number type new_col_number`
        where the operator can be \*, /, +, or -. Type indicates if the operation is between
        two columns (element-wise calculation) if : or if the operation is between a column
        and a scalar if =.
        """
        if not self._check_number_of_arguments(command, 8):
            return

        (_, arg_1, operation, arg_2, type_of_math,
         output_col_idx, output_name, output_unit) = command.split()

        arg_1 = int(arg_1)
        output_col_idx = int(output_col_idx)
        output_name = output_name.strip()
        output_unit = output_unit.strip()

        # The first arg is always a column, we get that data
        first_arg_data = self._get_data_by_index(arg_1)

        # Determine the second operand - column or scalar
        # If the type of math is : then it's an element by element operation between columns
        # If the type of math is = then it's a column and scalar operation
        if type_of_math == ':':
            arg_2 = int(arg_2)
            second_arg_data = self._get_data_by_index(arg_2)
        elif type_of_math == '=':
            arg_2 = float(arg_2)
            second_arg_data = arg_2
        else:
            self.command_invalid(command)

        # Determine the operation and do it
        if operation == '*':
            result = first_arg_data * second_arg_data
        elif operation == '/':
            result = first_arg_data / second_arg_data
        elif operation == '+':
            result = first_arg_data + second_arg_data
        elif operation == '-':
            result = first_arg_data - second_arg_data
        else:
            self.command_invalid(command)

        # Put the result back into the output
        self._set_data_by_index(output_col_idx, result)
        self._set_name_by_index(output_col_idx, output_name)
        self._set_units_by_index(output_col_idx, output_unit)

    def command_comment(self, command):
        """
        Process a comment line.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        In XLook, comments are simply ignored, but this function is provided should we find
        the need to log, or otherwise process them (potentially even for metadata extraction).
        """
        pass

    def command_begin(self, command):
        """
        Process the begin command.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The begin command, while essential in XLook is simply ignored here. It denoted the
        start of the file.
        """
        pass

    def command_com_file(self, command):
        """
        Process the com_file command.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The command com_file is ignored here, but indicated that this was a command file
        to XLook.
        """
        pass

    def command_invalid(self, command):
        """
        Process an invalid command.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        Issue a warning for invalid commands, but we keep going just like XLook did. Xlook
        printed a console message, but they were mostly ignored by users, the warning should
        at least grab attention.
        """
        warnings.warn(f'Invalid Command {command} - ignored and processing proceeding.')

    def _check_number_of_arguments(self, command, n_args):
        """
        Check that the command has the required number of arguments.

        Check that we have enough arguments (the command counts as one as well)
        and issue a warning if not.

        Parameters
        ----------
        command : str
            command from r file
        n_args : int
            number of arguments expected with the command (the command counts as one)

        Returns
        -------
        valid : boolean
            If the command has the correct number of arguments.

        Notes
        -----
        Returns False so we can bail on processing that command and keep running like
        XLook did.
        """
        n_args_received = len(command.split())
        if n_args_received != n_args:
            warnings.warn(f'Command {command} expected {n_args}, but received'
                          f' {n_args_received} - ignored and processing proceeding.')
            return False
        return True

    def command_summation(self, command):
        """
        Compute the cumulative sum of a column.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The XLook command is `summation column_number new_column_number`
        """
        if not self._check_number_of_arguments(command, 5):
            return
        (_, input_col_idx, output_col_idx, output_name, output_unit) = command.split()
        input_col_idx = int(input_col_idx)
        output_col_idx = int(output_col_idx)
        result = np.cumsum(self._get_data_by_index(input_col_idx))

        # Put the result back into the output
        self._set_data_by_index(output_col_idx, result)
        self._set_name_by_index(output_col_idx, output_name)
        self._set_units_by_index(output_col_idx, output_unit)

    def command_power(self, command):
        """
        Compute a column raised to a power.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The Xlook command is `power power_value column_number new_column_number`
        """
        if not self._check_number_of_arguments(command, 6):
            return
        (_, power, input_col_idx, output_col_idx,
         output_name, output_unit) = command.split()
        input_col_idx = int(input_col_idx)
        output_col_idx = int(output_col_idx)
        power = float(power)
        result = self._get_data_by_index(input_col_idx) ** power

        # Put the result back into the output
        self._set_data_by_index(output_col_idx, result)
        self._set_name_by_index(output_col_idx, output_name)
        self._set_units_by_index(output_col_idx, output_unit)

    def command_zero(self, command):
        """
        Zero a column at a record.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The Xlook command is `zero column_number record_index`.

        See Also
        --------
        pylook.calc.zero
        """
        if not self._check_number_of_arguments(command, 3):
            return
        (_, input_col_idx, zero_record) = command.split()
        input_col_idx = int(input_col_idx)
        zero_record = int(zero_record)
        result = lc.zero(self._get_data_by_index(input_col_idx) * units('dimensionless'),
                         zero_record)
        self._set_data_by_index(input_col_idx, result.m)  # We are not touching units and names

    def command_ec(self, command):
        """
        Perform a linear elastic correction of a column.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        This is not unit safe like the pure Python version would be as we're shedding units
        all around in the interpreter. The XLook command was `ec displacement_column_number
        load_column_number new_column_number first_row_index last_row_index`
        """
        if not self._check_number_of_arguments(command, 9):
            return

        (_, disp_col_idx, load_col_idx, output_col_idx,
         first_idx, last_idx, slope, output_name, output_unit) = command.split()
        disp_col_idx = int(disp_col_idx)
        load_col_idx = int(load_col_idx)
        output_col_idx = int(output_col_idx)
        first_idx = int(first_idx)
        last_idx = int(last_idx)
        slope = 1 / float(slope)

        # Get the data and assign units - they don't matter we just need them for the
        # pylook functions to work
        load_data = self._get_data_by_index(load_col_idx) * units('dimensionless')
        disp_data = self._get_data_by_index(disp_col_idx) * units('dimensionless')
        coeffs = [slope * units('dimensionless'), 0 * units('dimensionless')]

        ec_corrected_disp = disp_data - lc.elastic_correction(load_data, disp_data, coeffs)

        # Drop our unit charade
        ec_corrected_disp = ec_corrected_disp.m
        disp_data = disp_data.m

        disp_data[first_idx: last_idx] = ec_corrected_disp[first_idx: last_idx]
        self._set_data_by_index(output_col_idx, disp_data)
        self._set_name_by_index(output_col_idx, output_name)
        self._set_units_by_index(output_col_idx, output_unit)

    def command_r_col(self, command):
        """
        Remove a given column by setting it to empty and the name/units to None.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The Xlook command is `r_col column_number`
        """
        if not self._check_number_of_arguments(command, 2):
            return
        _, col_idx = command.split()
        col_idx = int(col_idx)
        # Set to empty and None for names and units
        self._set_data_by_index(col_idx, np.empty_like(self._get_data_by_index(col_idx)))
        self._set_name_by_index(col_idx, None)
        self._set_units_by_index(col_idx, None)

    def command_math_int(self, command):
        """
        Perform a math operation on an interval of a column.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The Xlook command is math_int x_col_number operator y_col_number type new_col_number
        first_row_index last_row_index`
        """
        if not self._check_number_of_arguments(command, 10):
            return

        (_, arg_1, operation, arg_2, type_of_math,
         output_col_idx, start_idx, stop_idx, output_name, output_unit) = command.split()

        arg_1 = int(arg_1)
        output_col_idx = int(output_col_idx)
        start_idx = int(start_idx)
        stop_idx = int(stop_idx)
        output_name = output_name.strip()
        output_unit = output_unit.strip()

        # The first arg is always a column, we get that data
        first_arg_data = self._get_data_by_index(arg_1)

        # Determine the second operand - column or scalar
        # If the type of math is : then it's an element by element operation between columns
        # If the type of math is = then it's a column and scalar operation
        if type_of_math == ':':
            arg_2 = int(arg_2)
            second_arg_data = self._get_data_by_index(arg_2)
        elif type_of_math == '=':
            arg_2 = float(arg_2)
            second_arg_data = arg_2
        else:
            self.command_invalid(command)

        # Determine the operation and do it
        if operation == '*':
            result = first_arg_data * second_arg_data
        elif operation == '/':
            result = first_arg_data / second_arg_data
        elif operation == '+':
            result = first_arg_data + second_arg_data
        elif operation == '-':
            result = first_arg_data - second_arg_data
        else:
            self.command_invalid(command)

        first_arg_data[start_idx: stop_idx] = result[start_idx: stop_idx]

        # Put the result back into the output
        self._set_data_by_index(output_col_idx, first_arg_data)
        self._set_name_by_index(output_col_idx, output_name)
        self._set_units_by_index(output_col_idx, output_unit)

    def command_offset_int(self, command):
        """
        Perform an offset over an interval in the data.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The Xlook command is `offset_int column_number record_start_index record_end_index
        (y or n) to offset in between during the offset.`
        """
        if not self._check_number_of_arguments(command, 5):
            return

        (_, col_idx, start_idx, stop_idx, set_between) = command.split()
        col_idx = int(col_idx)
        start_idx = int(start_idx)
        stop_idx = int(stop_idx)
        if set_between.strip().lower() == 'y':
            set_between = True
        elif set_between.strip().lower() == 'n':
            set_between = False
        else:
            self.command_invalid()

        col_data = self._get_data_by_index(col_idx)

        col_data = lc.remove_offset(col_data * units('dimensionless'), start_idx, stop_idx,
                                    set_between=set_between)

        self._set_data_by_index(col_idx, col_data.m)

    def command_r_row(self, command):
        """
        Remove a range of rows from all columns of data.

        Parameters
        ----------
        command : str
            command from r file

        Notes
        -----
        The Xlook command is `r_row column_number first_row_index last_row_index`.
        """
        if not self._check_number_of_arguments(command, 3):
            return
        _, start_row_idx, end_row_idx = command.split()
        start_row_idx = int(start_row_idx)
        end_row_idx = int(end_row_idx)
        if end_row_idx == -1:
            end_row_idx = None
        slice_to_delete = slice(start_row_idx, end_row_idx, None)
        for col_idx in range(self.max_cols):
            self._set_data_by_index(col_idx, np.delete(self._get_data_by_index(col_idx),
                                                       slice_to_delete))

    def command_read(self, command, path_relative_to_r_file=True, endianness=None):
        """
        Read a binary file in for processing.

        Parameters
        ----------
        command : str
            command from r file
        path_relative_to_r_file : boolean
            Determines if the path to be read is relative to the r file as xlook did or if
            it is relative to the calling Python code. Default True.

        Notes
        -----
        The Xlook command is `read filename`
        """
        if not self._check_number_of_arguments(command, 2):
            return
        _, fpath = command.split()
        fpath = Path(fpath.strip())

        if path_relative_to_r_file:
            fpath = self._r_file_path.parent / fpath.name

        data_dict, _ = read_binary(fpath, data_endianness=endianness)

        # Break the data dict out into the structure of the class
        for i, (name, data_col) in enumerate(data_dict.items()):
            self._set_data_by_index(i, data_col.m)
            self._set_name_by_index(i, name)
            self._set_units_by_index(i, str(data_col.units))

    def get_data_dict(self, data_units=None, ignore_unknown_units=False):
        """
        Format the data into a dictionary of quantity arrays.

        Create a data dict like the rest of pylook uses and attach user given units.
        If we get no user given units we try to parse what we have in Xlook and either
        error (default) or can assign unitless to everything that is unrecognized.

        Parameters
        ----------
        data_units : list
            List of quantities for each data column. Overwrites and units from the file
            metadata.
        ignore_unknown_units : boolean
            If True any units from the file metadata that we cannot parse are set to
            dimensionless and a warning issued. If False (default) an error is raised.

        Returns
        -------
        data : dict
            Dictionary of quantity arrays
        """
        # If units are not given, let's try to parse what we got from the units the user
        # gave Xlook.
        if data_units is None:
            data_units = self.data_units

        d = {}
        for i, (name, unit) in enumerate(zip(self.data_names, data_units)):

            # If there's nothing here, just to go to the next column
            if (name is None) and (unit is None):
                continue

            # XLook users commonly used a . for dimensionless
            if unit == '.':
                unit = 'dimensionless'

            try:
                d[name] = self.data[i] * units(unit)
            except UndefinedUnitError:
                if ignore_unknown_units:
                    d[name] = self.data[i] * units('dimensionless')
                    warnings.warn(f'Unknown unit {unit} for data {name}'
                                  ' was assigned dimensionless')
                else:
                    raise UndefinedUnitError(unit)
        return d



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Setup script for pylook."""

import sys

from setuptools import setup

if sys.version_info[0] < 3:
    error = """
    pylook requires Python 3.6 or greater!
    Python {py} detected.
    """.format(py='.'.join([str(v) for v in sys.version_info[:3]]))

    print(error)  # noqa: T001
    sys.exit(1)

setup(use_scm_version={'version_scheme': 'post-release'})



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Tools to process and reduce experimental rock mechanics data."""

import sys

if sys.version_info < (3,):
    raise ImportError(
        """
        You are running Python 2 which is a no longer supported version of Python and not
        compatible with pylook.
        """)

from ._version import get_version  # noqa: E402
__version__ = get_version()
del get_version
